from rich import print as rprint


def run_scrapper_with_success(folder_name: str):
	"""
	Print a success message after running the scrapper with success.
	"""

	rprint(
		"[bold white]Step :one:  [bold green] Done  :star::white_heavy_check_mark-emoji: ! [/bold green] [/bold white]  "
	)


def update_db_with_success():
	"""
	Print a success message after updating the database with new data.
	"""
	rprint("[bold white]Step :three:  is [bold green] done  :star: ! [/bold green] [/bold white]")
	rprint(
		"[bold white blink]Your database has been updated with new products and their related suppliers :white_heavy_check_mark-emoji: ![/bold white blink]"
	)
	rprint(
		"[bold white]Now you can do some amazing analysis with your database for making your business more profitable :star: :star: ! [/bold white ]"
	)


def update_db_success_sqlite(sqlite_file: str):
	"""Print a success message when the database has been updated with new products and suppliers."""
	rprint(
		f"[bold white][magenta bold]{sqlite_file}.sqlite [/magenta bold] file has been updated succesfully  with new products and their related suppliers :white_heavy_check_mark-emoji: ![/bold white]"
	)
	rprint("[bold white]Step :three:  is [bold green] done  :star: ! [/bold green] [/bold white]  ")
	rprint(
		"[bold white ]Now you can do some amazing analysis with your database for making your business more profitable :star: :star: ! [/bold white ]"
	)



import json
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import Sequence, Union
from selectolax.parser import Node

import rich
from loguru import logger
from rich.progress import (
	Progress,
	SpinnerColumn,
)
from selectolax.parser import HTMLParser

from aba_cli_scrapper.typed_datas import ProductDict, SupplierDict

from .html_to_disk import json_hunter
from .utils_scrapping import (
	country_name,
	custom_minium_to_oder,
	get_product_certification,
	get_product_price,
	is_alibaba_guaranteed,
	is_customizable,
	is_full_promotion,
	is_instant_order,
	is_trade_product,
	ordered_or_sold,
	suppliers_status,
)


@dataclass
class PageParser:
	"""Parse each html page downloaded from alibaba and return a list of suppliers and products.

	:raises TypeError: _description_
	:raises FileNotFoundError: if the targeted folder does not exist
	:return: PageParser object
	:rtype: class `PageParser`
	"""

	# folder path that contains all html files to be scraped
	targeted_folder: Union[str, Path]

	def _retrieve_html_content_as_string(self, html_file: Path):
		"""Retrieve html content from a html file and return it as a string.

		:param html_file: The path of the html file to be read.
		:type html_file: str
		:return: The html content as a string.
		:rtype: str
		"""
		logger.info(f"Retrieving html content from file : {html_file}")
		html_content = html_file.read_text(
			encoding="utf-8",
		)
		logger.info("Html content has been retrieved.")
		return html_content

	@logger.catch(FileNotFoundError)
	def _html_files_explorer(self):
		"""
		Retrieves all the html files in the targeted folder.

		:raises FileNotFoundError: if the targeted folder does not exist
		:return: A list of all html files in the targeted folder
		:rtype: List[Path]
		"""
		targeted_folder = Path(self.targeted_folder).resolve()
		if not targeted_folder.exists():
			raise FileNotFoundError()
		logger.info(f"getting html files from {targeted_folder} ... ")
		html_files = [html_file for html_file in targeted_folder.glob("*.html")]
		logger.info("Html files list has been returned.")
		return html_files

	def _divs_and_dict(self, selector: str = ".fy23-search-card"):
		"""
		Retrieves the div tags that match the given selector and a dictionary of parsed content from the html file
		containing the div tag.

		:param selector: The css selector to retrieve the div tags.
		:type selector: str
		:return: A list of tuples containing the div tag and its corresponding parsed content.
		:rtype: List[Tuple[selectolax.Selector, Dict]]
		"""
		logger.info(f"Retrieving  div tags with class='{selector}' ...")
		divs_and_dict = [
			(
				HTMLParser(self._retrieve_html_content_as_string(html_file)).css(selector),
				json_hunter(
					self._retrieve_html_content_as_string(html_file),
					css_selector="div[id='root']",
				),
			)
			for html_file in self._html_files_explorer()
		]
		logger.info("All expected divs and dict has been retrived ...")

		new_divs_and_dict = []
		for item in divs_and_dict:
			if item[1] is None or item[1] == "":
				continue
			new_divs_and_dict.append((item[0], json.loads(item[1])))  # type: ignore
		return new_divs_and_dict

	def _suppliers_appender(self, offers_list: Sequence[dict], suppliers: list, divs: list[Node]):
		for offer in offers_list:
			suppliers.append({
				"name": offer["companyName"].lower(),
				"verified_type": suppliers_status(tags=divs, offer=offer),
				"sopi_level": offer["displayStarLevel"],
				"country_name": country_name(country_min=offer["countryCode"]),
				"gold_supplier_year": offer["goldSupplierYears"].split(" ")[0],
				"supplier_service_score": float(offer["supplierService"]),
			})
		return suppliers

	def detected_suppliers(self):
		"""Retrieves the detected suppliers from the divs and offers data.
		Returns a list of unique suppliers with their relevant information.
		"""
		suppliers: Sequence[SupplierDict] = list()
		console = rich.console.Console()
		with Progress(
			SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"),
			*Progress.get_default_columns(),
			console=console,
			transient=True,
		) as progress:
			all_pages_task = progress.add_task(
				description="[blank]Parsing suppliers ...[/blank]",
			)
			# single_page_task = progress.add_task(description="Retrieving suppliers ...", total=1)
			for divs, offers in self._divs_and_dict():
				with open("offers_i.json", "w") as f:
					f.write(json.dumps(offers, indent=4))
				try:
					if offers["offerTotalCount"] == 0:
						continue
				except KeyError:
					if offers["props"]["offerTotalCount"] == 0:
						continue
				try:
					offers_values = offers["offerResultData"]["offers"]
					suppliers_added = self._suppliers_appender(
						offers_list=offers_values, suppliers=suppliers, divs=divs
					)
				except KeyError:
					offers_values = offers["props"]["offerResultData"]["offers"]
					suppliers_added = self._suppliers_appender(
						offers_list=offers_values, suppliers=suppliers, divs=divs
					)
				progress.update(all_pages_task, advance=100 / len(self._divs_and_dict()))
		# removing supliers present twice in supliers dict
		unique_suppliers_tuple = list(OrderedDict((str(d), d) for d in suppliers_added).items())
		unique_suppliers = [supplier_tuple[1] for supplier_tuple in unique_suppliers_tuple]
		return unique_suppliers

	def _produtcs_appender(self, offers_list: Sequence[dict], products: Sequence[ProductDict]):
		for offer in offers_list:
			products.append({
				"name": offer["enPureTitle"].lower(),
				"max_price": get_product_price(all_price_text=offer["price"], which="max"),
				"min_price": get_product_price(all_price_text=offer["price"], which="min"),
				"guaranteed_by_alibaba": is_alibaba_guaranteed(str_status=offer["halfTrust"]),
				"certifications": get_product_certification(offer=offer),
				"minimum_to_order": custom_minium_to_oder(offer["halfTrustMoq"].lower()),
				"ordered_or_sold": ordered_or_sold(offer=offer),
				"supplied_by": offer["companyName"].lower(),
				"product_score": float(offer["productScore"]),
				"review_count": float(offer["reviewCount"]),
				"review_score": float(offer["reviewScore"]),
				"shipping_time_score": float(offer["shippingTime"]),
				"is_full_promotion": is_full_promotion(str_status=offer["isFullPromotion"]),
				"customizable": is_customizable(str_status=offer["customizable"]),
				"instant_order": is_instant_order(str_status=offer["halfTrustInstantOrder"]),
				"trade_product": is_trade_product(str_status=offer["tradeProduct"]),
			})
		return products

	def detected_products(self):
		"""Returns a list of unique products with their relevant information."""
		products: Sequence[ProductDict] = list()
		console = rich.console.Console()
		with Progress(
			SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"),
			*Progress.get_default_columns(),
			console=console,
			transient=True,
		) as progress:
			all_pages_task = progress.add_task(description="[blank]Parsing products ...[/blank]")
			for divs, offers in self._divs_and_dict():
				try:
					if offers["offerTotalCount"] == 0:
						continue
				except KeyError:
					if offers["props"]["offerTotalCount"] == 0:
						continue
				try:
					offers_values = offers["offerResultData"]["offers"]
					products_added = self._produtcs_appender(
						offers_list=offers_values, products=products
					)
				except KeyError:
					offers_values = offers["props"]["offerResultData"]["offers"]
					products_added = self._produtcs_appender(
						offers_list=offers_values, products=products, divs=divs
					)
				progress.update(all_pages_task, advance=100 / len(self._divs_and_dict()))
		unique_products_tuple = list(OrderedDict((str(d), d) for d in products_added).items())
		# removing supliers present twice in supliers dict
		unique_products = [product_tuple[1] for product_tuple in unique_products_tuple]
		return unique_products



"""This script is responsible for scraping data from the Alibaba website.
It uses the Playwright library to navigate through multiple pages
extract HTML content,and save it to disk. The script defines
functions to handle asynchronous tasks,interact with
browser contexts, and gather data from the scraped HTML files.
Additionally, it utilizes loguru for logging purposes and decouple for
managing environment variables.
"""

import asyncio
import os
import platform
import urllib3

from click import UsageError
from loguru import logger
from playwright.async_api import async_playwright, APIRequestContext
from playwright.sync_api import Error as PError
from playwright.sync_api import sync_playwright
from rich import print as rprint
from rich.progress import Progress, SpinnerColumn, TaskID
import typer

from . import BRIGHT_DATA_API_KEY, SYPHOON_API_KEY
from .html_to_disk import write_to_disk
from .info_message import run_scrapper_with_success
from .proxies_utils import urls_pusher, goto_task, HTML_PAGE_RESULT


class BrightDataProxyProvider:
	BD_API_KEY = BRIGHT_DATA_API_KEY

	@classmethod
	@logger.catch()
	async def async_scraper(cls, *, save_in: str, key_words: str, page_results: int) -> None:
		"""
		Initiates scraping of Alibaba.com based on the provided keywords.

		:param save_in: The directory to store the raw HTML files.
		:type save_in: str
		:param key_words: The search term(s) for finding products on Alibaba. Enclose multiple keywords in quotes.
		:type key_words: str
		:param page_results: The number of pages to scrape. If omitted, 10 pages will be scraped.
		:type page_results: int

		:return: None
		:rtype: None

		:raises UsageError: If no internet connection is available.

		:raises SystemExit: If the Bright data account has been suspended by the system.

		:raises SystemExit: If no SCRAPING BROWSER API key is provided.

		:raises playwright._impl._errors.Error: If there is an error with the playwright browser.

		:raises Exception: If there is an error processing a page.
		"""
		if cls.BD_API_KEY == "":
			rprint("[red]You need to set your  API key to use BrightData proxies ... [/red]")
			return typer.Exit(code=1)
		with Progress(
			SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"),
			*Progress.get_default_columns(),
			transient=True,
		) as progress:
			async with async_playwright() as p:
				logger.info("Connecting to CDP and creating the browser... ")
				try:
					# print("rt")
					if cls.BD_API_KEY == "":
						rprint(
							"[red]You need to set your SCRAPING BROWSER API key from BrightData to Enable Async Scraping"
						)
						return
					browser = await p.chromium.connect_over_cdp(cls.BD_API_KEY)
				except urllib3.exceptions.NameResolutionError:
					rprint("[red]check your internet connection")
					return
				except Exception as e:
					if "Account is suspended" in str(e):
						rprint(
							"[red]Bright data account has been suspended by system may your api is exhausted recharge it and try again. You can use `--sync-api` flag after your last command to enable sync scrapping but you may not encounter enought success.  [/red]"
						)
						return
					elif "Executable doesn't exist at" in str(e):
						rprint(
							"[white]Seems like playwright is not installed or needs to be update. We will install it for you after that rerun your previous command [/white]"
						)
						os.system("playwright install")
						if platform.system() == "Windows":
							os.system("cls")
						if platform.system() == "Linux":
							os.system("clear")
					elif "WebSocket error" in str(e):
						rprint(
							"[red]Web Socket is disconnected. You May need to activate your Internet connexion"
						)
						return typer.Exit(code=1)
					elif "try connecting via ws" in e.message:
						rprint(
							"[red] You must to set a SCRAPING BROWSER API key (e.g 'ws://127.0.0.1:3000') to enable connection via websocket. "
						)
						return typer.Exit(code=1)
					else:
						rprint(f"[red]Unexpected error occured : \n{e}")
						return typer.Exit(code=1)

				context_browser = await browser.new_context()
				s_one = asyncio.Semaphore(value=10)
				logger.info("Loading pages results ... ")
				async with asyncio.Lock():
					async with asyncio.TaskGroup() as tg:
						task = progress.add_task("[green blink] async Scraping...", start=False)
						for url in urls_pusher(words=key_words, stop_at=page_results):
							async with asyncio.Lock():
								page = await context_browser.new_page()
							tg.create_task(
								goto_task(
									url=url,
									semaphore=s_one,
									progress=progress,
									task=task,
									tg_instance=tg,
									context_browser=context_browser,
									page=page,
									page_results=page_results,
								)
							)
				write_to_disk(save_in, HTML_PAGE_RESULT)
				run_scrapper_with_success(folder_name=save_in)

	@classmethod
	def sync_scraper(cls, *, save_in: str, key_words: str, page_results: int) -> None:
		if cls.BD_API_KEY == "":
			rprint("[red]You need to set your  API key to use BrightData proxies ... [/red]")
			return typer.Exit(code=1)
		with Progress(
			SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"),
			*Progress.get_default_columns(),
			transient=True,
		) as progress:
			task = progress.add_task(
				"[green blink] Sync Scraping...",
				start=False,
			)
			playwright = sync_playwright().start()
			try:
				browser = playwright.chromium.connect_over_cdp(cls.BD_API_KEY)
			except PError as e:
				if "exists" in e.message:
					rprint(
						"[white] Seems like playwright is not installed or need to be update. lets aba install it for you... [/white]"
					)
					os.system("playwright install")
					return typer.Exit(code=1)
				elif "try connecting via ws" in e.message:
					rprint(
						"[red] You must to set a SCRAPING BROWSER API key (e.g 'ws://127.0.0.1:3000') to enable connection via websocket. "
					)
					return typer.Exit(code=1)
				else:
					rprint(f"[red]Unexpected error occured : \n{e}")
					return typer.Exit(code=1)
			context = browser.new_context()
			for url in urls_pusher(words=key_words, stop_at=page_results):
				logger.info(f"Loading page {url.split('page=')[1]} ... ")
				page = context.new_page()
				try:
					response = page.goto(url, wait_until="domcontentloaded", timeout=0)
					if response is None:
						return None
					logger.info(
						f"Returns the text representation of response body from page {url.split('page=')[1]} ... "
					)
					html_content = response.text()
					progress.start_task(task)
					progress.update(task, advance=100 / page_results)
					global HTML_PAGE_RESULT
					HTML_PAGE_RESULT.append(html_content)
					logger.info(f"Closing the page {url.split('page=')[1]} ... ")
					page.close()
				except PError as e:
					if "ERR_INTERNET_DISCONNECTED" in e.message:
						raise UsageError("Check your internet connection ... ")
			write_to_disk(save_in, HTML_PAGE_RESULT)
			run_scrapper_with_success(folder_name=save_in)


class SyphoonProxyProvider:
	SP_API_KEY = SYPHOON_API_KEY

	@classmethod
	@logger.catch()
	async def custom_goto_task(
		cls,
		url: str,
		api_context: APIRequestContext,
		task: TaskID,
		progress: Progress,
		semaphore: asyncio.Semaphore,
		tg_instance: asyncio.TaskGroup,
		page_results: int,
	):
		logger.info(f"fetching page {url.split('page=')[1]} ... ")
		response = await api_context.post(
			"http://api.syphoon.com/",
			data={
				"key": cls.SP_API_KEY,
				"method": "get",
				"url": url,
			},
			timeout=0,
		)
		if response is None:
			return None
		logger.info(f"get response from syphoon api for: {url.split('page=')[1]} ... ")
		progress.start_task(task)
		html_body = await response.text()
		progress.update(task, advance=100 / page_results)
		global HTML_PAGE_RESULT
		HTML_PAGE_RESULT.append(html_body)
		logger.info(f"Closing page {url.split('page=')[1]} ... ")
		return

	@classmethod
	@logger.catch()
	def sync_scraper(cls, *, save_in: str, key_words: str, page_results: int) -> None:
		if cls.SP_API_KEY == "":
			rprint("[red]You need to set your  API key to use Syphoon proxies ... [/red]")
			return typer.Exit(code=1)
		with Progress(
			SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"),
			*Progress.get_default_columns(),
			transient=True,
		) as progress:
			task = progress.add_task(
				"[green blink] Sync Scraping...",
				start=False,
			)
			playwright = sync_playwright().start()
			try:
				browser = playwright.chromium.launch(headless=True)
			except PError:
				rprint(
					"[white] Seems like playwright is not installed or needs to be update. lets aba install it for you... [/white]"
				)
				os.system("playwright install")
				return typer.Exit(code=1)
			context = browser.new_context()
			api_request = context.request
			for url in urls_pusher(words=key_words, stop_at=page_results):
				logger.info(f"Loading page {url.split('page=')[1]} ... ")
				try:
					response = api_request.post(
						"http://api.syphoon.com/",
						data={
							"key": cls.SP_API_KEY,
							"method": "get",
							"url": url,
						},
						timeout=0,
					)
					if response is None:
						return None
					logger.info(
						f"Returns the text representation of response body from page {url.split('page=')[1]} ... "
					)
					progress.start_task(task)
					html_content = response.text()
					progress.update(task, advance=100 / page_results)
					global HTML_PAGE_RESULT
					HTML_PAGE_RESULT.append(html_content)
					logger.info(f"Closing the page {url.split('page=')[1]} ... ")
				except PError as e:
					if "ERR_INTERNET_DISCONNECTED" in e.message:
						raise UsageError("Check your internet connection ... ")
		write_to_disk(save_in, HTML_PAGE_RESULT)
		run_scrapper_with_success(folder_name=save_in)

	@classmethod
	@logger.catch()
	async def async_scraper(cls, *, save_in: str, key_words: str, page_results: int) -> None:
		"""
		Initiates scraping of Alibaba.com based on the provided keywords.

		:param save_in: The directory to store the raw HTML files.
		:type save_in: str
		:param key_words: The search term(s) for finding products on Alibaba. Enclose multiple keywords in quotes.
		:type key_words: str
		:param page_results: The number of pages to scrape. If omitted, 10 pages will be scraped.
		:type page_results: int

		:return: None
		:rtype: None

		:raises UsageError: If no internet connection is available.

		:raises SystemExit: If the Bright data account has been suspended by the system.

		:raises SystemExit: If no SCRAPING BROWSER API key is provided.

		:raises playwright._impl._errors.Error: If there is an error with the playwright browser.

		:raises Exception: If there is an error processing a page.
		"""
		browser = None
		if cls.SP_API_KEY == "":
			rprint("[red]You need to set your  API key to use Syphoon proxies ... [/red]")
			return typer.Exit(code=1)
		with Progress(
			SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"),
			*Progress.get_default_columns(),
			transient=True,
		) as progress:
			async with async_playwright() as p:
				logger.info("Connecting to CDP and creating the browser... ")
				try:
					if cls.SP_API_KEY == "":
						rprint(
							"[red]You need to set your SCRAPING BROWSER API key from BrightData or proxies API key from Syphoon"
						)
						return typer.Exit(code=1)
					browser = await p.chromium.launch()
				except PError as e:  # type: ignore
					if "Account is suspended" in str(e):
						rprint(
							"[red]Bright data account has been suspended by system may your api is exhausted recharge it and try again. You can use `--sync-api` flag after your last command to enable sync scrapping but you may not encounter enought success.  [/red]"
						)
						return typer.Exit(code=1)
					elif "WebSocket error" in str(e):
						rprint(
							"[red]Web Socket is disconnected. You May need to activate your Internet connexion"
						)
						return typer.Exit(code=1)
					elif "Executable doesn't exist at" in str(e):
						rprint(
							"[white]Seems like playwright is not installed or needs to be update. We will install it for you after that rerun your previous command [/white]"
						)
						os.system("playwright install")
						if platform.system() == "Windows":
							os.system("cls")
						if platform.system() == "Linux":
							os.system("clear")
					else:
						rprint(f"[red]Unexpected error occured ...:{str(e)}[/red]")
						return typer.Exit(code=1)
				context_browser = await browser.new_context(base_url="http://api.syphoon.com/")
				s_one = asyncio.Semaphore(value=2)
				logger.info("Loading pages results ... ")
				api_request_context = context_browser.request
				async with asyncio.Lock():
					async with asyncio.TaskGroup() as tg:
						task = progress.add_task("[green blink] async Scraping...", start=False)
						for url in urls_pusher(words=key_words, stop_at=page_results):
							async with asyncio.Lock():
								tg.create_task(
									cls.custom_goto_task(
										url=url,
										api_context=api_request_context,
										task=task,
										progress=progress,
										semaphore=s_one,
										tg_instance=tg,
										page_results=page_results,
									)
								)
				write_to_disk(save_in, HTML_PAGE_RESULT)
				run_scrapper_with_success(folder_name=save_in)
